import hashlib
import pathlib
from rich.progress import track
from rich import print
from rich.traceback import install
from datetime import datetime

import click
import html2text
import sqlite_utils

install(show_locals=False)


FILES_TABLE = "files"


class TableColumns(object):
    ID = "id"
    NAME = "name"
    SIZE = "size"
    PATH = "path"
    ADDED = "added"
    PLAINTEXT_CONTENT = "plaintext-content"
    CONTENT = "content"


def html_file_to_text(file):
    with open(file, "r") as file_contents:
        return html2text.html2text(file_contents.read())


def initialize_db(db: sqlite_utils.Database):
    schema = db.schema
    if not schema:
        print(f"Creating database schema...")
        table = db[FILES_TABLE]
        table.create(
            {
                TableColumns.ID: str,
                TableColumns.NAME: str,
                TableColumns.SIZE: int,
                TableColumns.PATH: str,
                TableColumns.ADDED: str,
                TableColumns.PLAINTEXT_CONTENT: str,
                TableColumns.CONTENT: bytes,
            },
            pk="id",
        )
        table.enable_fts(
            ["name", "plaintext-content"], create_triggers=True, tokenize="porter"
        )


@click.argument("input_dirs", type=click.Path(file_okay=False, exists=True), nargs=-1)
@click.option(
    "--database",
    "-d",
    type=click.Path(dir_okay=False),
    default="htmlstore.db",
    help="The index database to update.",
)
@click.option(
    "--store-binary",
    "-b",
    is_flag=True,
    default=False,
    help="Insert the file's contents into SQLite DB as binary.",
)
@click.command()
def index(input_dirs: str, database: str, store_binary: bool):
    """Indexes the HTMLs files so they can be searched."""

    print(f"Using database [bold]{database}[/bold].")
    db = sqlite_utils.Database(database)
    initialize_db(db)

    for input_dir in input_dirs:
        search_path = pathlib.Path(input_dir)
        files_to_process = [
            file
            for file in search_path.iterdir()
            if file.is_file() and file.suffix.lower() in [".html", ".htm"]
        ]

        files_indexed = 0
        files_not_indexed = 0
        for file in track(
            files_to_process, description=f'Indexing files on "{search_path}"...'
        ):
            path = str(file)
            previous_file_entry = next(
                db.query(
                    f"select id from {FILES_TABLE} where {TableColumns.PATH} = :{TableColumns.PATH}",
                    {TableColumns.PATH: path},
                ),
                None,
            )
            if not previous_file_entry:
                files_indexed = files_indexed + 1
                filename = file.name
                id_ = hashlib.md5(path.encode("UTF-8")).hexdigest()
                size = file.stat().st_size
                with open(file, "r") as file_handler:
                    content = file_handler.read()
                    values = {
                        TableColumns.ID: id_,
                        TableColumns.NAME: filename,
                        TableColumns.SIZE: size,
                        TableColumns.PATH: path,
                        TableColumns.ADDED: datetime.now().strftime('%d-%m-%Y'),
                        TableColumns.PLAINTEXT_CONTENT: html2text.html2text(content),
                    }
                    if store_binary:
                        values[TableColumns.CONTENT] = content
                    db["files"].insert(values, pk="id")
            else:
                files_not_indexed = files_not_indexed + 1

        if files_not_indexed:
            print(f"{files_not_indexed} file(s) were already indexed!")
        if files_indexed:
            print(f"{files_indexed} file(s) were indexed.")


@click.option(
    "--database",
    "-d",
    type=click.Path(dir_okay=False),
    default="htmlstore.db",
    help="The index database to update.",
)
@click.option("--dry-run", is_flag=True, help="Do not perform any actions.")
@click.command()
def purge(database: str, dry_run: bool):
    """Removes all files which content is not accessible (on the filesystem or on the database)."""

    print(f"Using database [bold]{database}[/bold].")
    db = sqlite_utils.Database(database)
    initialize_db(db)
    files_table = db[FILES_TABLE]

    missing_files_found = 0
    for row in db.query(f"select id, path from {FILES_TABLE} where content is null"):
        path = pathlib.Path(row[TableColumns.PATH])
        if not path.exists():
            id_ = row[TableColumns.ID]
            missing_files_found = missing_files_found + 1
            if not dry_run:
                print(f"Removing entry [bold]{id_}[/bold] [blue]{path}[/blue]...")
                files_table.delete(id_)
            else:
                print(f"Entry [bold]{id_}[/bold] [blue]{path}[/blue] would be removed.")

    if missing_files_found == 0:
        print("All files were accessible!")
    else:
        print(f"{missing_files_found} file(s) were not accessible!")


@click.option(
    "--output",
    "-o",
    type=click.Path(
        file_okay=False,
        exists=True,
    ),
    help="Specify an output directory. If defined DB entries will be updated if needed with the new location.",
)
@click.option(
    "--database",
    "-d",
    type=click.Path(dir_okay=False),
    default="htmlstore.db",
    help="The index database to update.",
)
@click.option("--dry-run", is_flag=True, help="Do not perform any actions.")
@click.command()
def extract(output: str, database: str, dry_run: bool):
    """Extracts all entries whose binary content is stored in the database to files"""

    print(f"Using database [bold]{database}[/bold].")
    db = sqlite_utils.Database(database)
    initialize_db(db)
    files_table = db[FILES_TABLE]
    output_path = pathlib.Path(output) if output else None

    warning_files_found = 0
    files_with_parent_directory_not_found = []
    files_to_extract = []

    missing_files_found = 0
    for row in db.query(
        f"select id, path from {FILES_TABLE} where content is not null"
    ):
        path = pathlib.Path(row[TableColumns.PATH])
        if (not output_path and path.exists()) or (output_path and output_path.joinpath(path.name).exists()):
            warning_files_found = warning_files_found + 1
        elif not output_path and not path.parent.exists():
            files_with_parent_directory_not_found.append(row)
        else:
            files_to_extract.append(row)

    if warning_files_found > 0:
        print(
            f"{warning_files_found} Files were [magenta]already found[/magenta] on their destinations."
        )

    parent_directory_not_found = set(
        [
            pathlib.Path(x[TableColumns.PATH]).parent
            for x in files_with_parent_directory_not_found
        ]
    )
    if parent_directory_not_found:
        print(
            f"{len(parent_directory_not_found)} Output directories [magenta]do not exist[/magenta]!"
        )
        for dir in parent_directory_not_found:
            print(f" - {dir}")
        return

    for row in track(files_to_extract, description="Extracting files..."):
        content = next(db.query(
            f"select {TableColumns.CONTENT} from {FILES_TABLE} where {TableColumns.ID} = :id",
            {"id": row[TableColumns.ID] }
        ))[TableColumns.CONTENT]
        path = output_path.joinpath(pathlib.Path(row[TableColumns.PATH]).name) if output_path else row[TableColumns.PATH]
        with open(path, 'w') as file:
            file.write(content)


@click.group()
def main():
    """Indexes HTML files to allow perform full text search over them."""
    pass


# Add commands and run
def cli():
    main.add_command(index)
    main.add_command(purge)
    main.add_command(extract)
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['htmls_to_datasette']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.1,<9.0.0',
 'datasette-render-html>=0.1.2,<0.2.0',
 'html2text>=2020.1.16,<2021.0.0',
 'rich>=10.10.0,<11.0.0',
 'sqlite-utils>=3.17,<4.0']

entry_points = \
{'console_scripts': ['htmls-to-datasette = htmls_to_datasette.cli:cli']}

setup_kwargs = {
    'name': 'htmls-to-datasette',
    'version': '0.1.1',
    'description': 'Tool to index and serve HTML files. Powered by Datasette.',
    'long_description': "# htmls-to-datasette\n\nHtmls-to-datasette is a tool to index HTML files into a [Sqlite](https://sqlite.org) database so they can be searched and\nvisualized at a later time. This can be useful for web archival/web clipping purposes.\n\nThe database created is designed to be served on [Datasette](https://datasette.io/) and to allow to read the indexed\nfiles through it. \n\nThis tool was created to serve my own work flow that is:\n 1. Have a browser with [SingleFile](https://github.com/gildas-lormeau/SingleFile) extension installed.\n 2. When there is an interesting blog post or article save a full web page into one HTML using SingleFile.\n 3. The created `.html` file on the downloads folder is moved to a common repository (via cron job).\n 4. This common repository is synched to my main server (I use [Syncthing](https://syncthing.net/) for this).\n 5. On my personal server all the new HTML files are moved to the serving folder and this indexer is called to populate\n    the search database.\n 6. Datasette with an specific configuration will allow searching on these files and reading them online.\n\nThe indexing tool can insert the HTML contents on the database itself, to be served from there, or not. In this second\ncase the files will be served from the location they were indexed from. \n\n## Setup\n\nThis project uses *[Poetry](https://python-poetry.org/)* to make it easier to setup the appropriate dependencies to run.\n\nInstallation steps for *Poetry* can be checked on their [website](https://python-poetry.org/docs/#installation) but for\nmost of the cases this command line would work:\n```bash\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python -\n```\n*Note that you should exercise caution when running something directly from Internet.*  \n\n### Install dependencies:\n```bash\npoetry install\n```\n\n### Build\n\n```bash\npoetry build\n```\n\n### Install\n\nI use [pipx](https://pypa.github.io/pipx) for installing packages on isolated environments. You can install this package\nfrom the `dist/` directory in whichever way you prefer or you can \n[install pipx](https://pypa.github.io/pipx/installation/).  \n\nThe installation with pipx would be:\n```bash\npipx install dist/htmls-to-datasette-0.1.0.tar.gz\n```\n\n## Usage\n\n`htmls-to-datasette index [OPTIONS] [INPUT_DIRS]...` will create a database named `htmlstore.db' (by default).\n\n### Example\n\nGet into the server directory:\n```bash\ncd server\n```\n\nBecause this example requires Datasette to run you would have to get them using poetry:\n```bash\npoetry init\n```\n\nNow index the example file using `htmls-to-datasette`:\n```bash\nhtmls-to-datasette index input\n```\n\nAll files contained in `input` (`.html` and `.html`) will be indexed and a full text search index created. Whenever\nthere are new files to be indexed this command can be run in the same way.\n\nAnd now run the Datasette server:\n```bash\npoetry run datasette serve htmlstore.db -m metadata-files.json --plugins-dir=plugins\n```\n\nYou'll see the address to send your browser to on the screen. There is also a shortcut to make it easier to perform a\nfull text search. Should be reachable at http://127.0.0.1:8001/htmlstore/search just fill the query on the 'q' parameter\nand you will search over the indexed HTMLs. Click on the HTML file name will load its contents.\n\nFor this to work the server will require the files to be on their location (relative in this case). So if the `input`\nfolder is moved away or not accesible the files would be searchable but the contents will not be available.\n\nThere is an additional example that stores these files onto the Sqlite database itself. This has its advantages as\neverything needed for serving and searching the content will be contained in one file.\n\n```bash\n# You should be on the server directory\nrm htmlstore.db   # Remove the previous example's database\nhtmls-to-datasette index input --store-binary  # Index files and store its contents\n\n# Now run Datasette, note that now we need to use a different metadata as the contents needed to be served\n# in a different way (from the DB itself). \npoetry run datasette serve htmlstore.db -m metadata-binary.json --plugins-dir=plugins\n```\n\n### TODO\n\n- Clear content when extracting files.\n- Better documentation.\n",
    'author': 'Pablo Lopez-Jamar',
    'author_email': 'pablo.lopez.jamar@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pjamar/htmls-to-datasette',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
